@@ -117,6 +117,8 @@ def ixs_thatsort_a2b(a: np.ndarray, b: np.ndarray, check_content: bool = True) -
 	return np.argsort(a)[np.argsort(np.argsort(b))]
 
 
+# TODO: This function is a copy of the same function in loompy.metadata_loaders, call that one instead
+
 def load_sample_metadata(path: str, sample_id: str) -> Dict[str, str]:
 	if not os.path.exists(path):
 		raise ValueError(f"Samples metadata file '{path}' not found.")
@@ -325,11 +327,11 @@ class BusFile:
 		row_attrs = {}
 		# Transpose the gene metadata
 		for i, attr in enumerate(self.gene_metadata_attributes):
-			row_attrs[attr] = [v[i] for v in self.genes.values()]
+			row_attrs[attr] = np.array([v[i] for v in self.genes.values()])
 
 		# Create cell attributes
 		col_attrs = {
-			"CellID": [sample_id + "_" + twobit_to_dna(int(cid), 16) for cid in self.cell_ids],
+			"CellID": np.array([sample_id + "_" + twobit_to_dna(int(cid), 16) for cid in self.cell_ids]),
 			"TotalUMIs": self.total_umis[self.valid_cells]
 		}
 
