@@ -25,6 +25,7 @@
 import gzip
 import logging
 import os.path
+import re
 from shutil import copyfile
 from typing import Tuple, Union, Any, Dict, List, Iterable, Callable, Optional
 
@@ -33,9 +34,12 @@ import numpy as np
 import numpy_groupies.aggregate_numpy as npg
 import scipy.sparse
 from scipy.io import mmread
+import scipy.sparse as sparse
 
 import loompy
 from loompy import deprecated, timestamp
+from loompy.metadata_loaders import make_row_attrs_from_gene_metadata, load_sample_metadata
+from .cell_calling import call_cells
 
 
 class LoomConnection:
@@ -79,7 +83,7 @@ class LoomConnection:
 		if validate:
 			lv = loompy.LoomValidator()
 			if not lv.validate(filename):
-				raise ValueError("\n".join(lv.errors) + f"\n{filename} does not appead to be a valid Loom file according to Loom spec version '{lv.version}'")
+				raise ValueError("\n".join(lv.errors) + f"\n{filename} does not appear to be a valid Loom file according to Loom spec version '{lv.version}'")
 
 		self._file = h5py.File(filename, mode)
 		self._closed = False
@@ -316,6 +320,7 @@ class LoomConnection:
 				self.shape = (self.ra[k].shape[0], self.shape[1])
 			if len(self.ca) == 0:
 				for k, v in col_attrs.items():
+					v = np.array(v)
 					self.ca[k] = np.zeros(0, v.dtype)
 
 		layers_dict: Dict[str, np.ndarray] = {}
@@ -388,6 +393,7 @@ class LoomConnection:
 			self.shape = (self.shape[0], n_cols)
 			todel = []
 			for key, vals in col_attrs.items():
+				vals = np.array(vals)
 				if vals.shape[1:] != self.col_attrs[key].shape[1:]:
 					logging.debug(f"Removing attribute {key} because shape {vals.shape} did not match existing shape {self.col_attrs[key].shape} beyond first dimension")
 					todel.append(key)
@@ -892,6 +898,8 @@ class LoomConnection:
 						if np.issubdtype(type(val), np.str_):
 							valnew = val.replace("/", "-")  # Slashes are not allowed in attribute names
 							valnew = valnew.replace(".", "_")  # Nor are periods
+						else:
+							valnew = val
 						ca[key + "_" + str(valnew)] = npg.aggregate(zero_strt_sort_noholes_lbls, (self.ca[key] == val).astype('int'), func="sum", fill_value=0)
 				elif func == "mode":
 					def mode(x):  # type: ignore
@@ -1048,10 +1056,10 @@ def create(filename: str, layers: Union[np.ndarray, Dict[str, np.ndarray], loomp
 			raise ValueError(f"Layer '{name}' is not the same shape as the main matrix")
 	for name, ra in row_attrs.items():
 		if len(ra) != shape[0]:
-			raise ValueError(f"Row attribute '{name}' is not the same length ({ra.shape[0]}) as number of rows in main matrix ({shape[0]})")
+			raise ValueError(f"Row attribute '{name}' is not the same length ({len(ra)}) as number of rows in main matrix ({shape[0]})")
 	for name, ca in col_attrs.items():
 		if len(ca) != shape[1]:
-			raise ValueError(f"Column attribute '{name}' is not the same length ({ca.shape[0]}) as number of columns in main matrix ({shape[1]})")
+			raise ValueError(f"Column attribute '{name}' is not the same length ({len(ca)}) as number of columns in main matrix ({shape[1]})")
 
 	try:
 		with new(filename, file_attrs=file_attrs) as ds:
@@ -1195,6 +1203,124 @@ def create_from_matrix_market(out_file: str, sample_id: str, layer_paths: Dict[s
 				ds[name] = layer
 
 
+def create_from_star(indir : str, outfile : str, sample_id : str, \
+                     cell_filter : str = "star", expected_n_cells : int = 0, min_total_umis : int = 0, ambient_pthreshold : float = 1.0, \
+                     sample_metadata_file : str = None, gtf_file : str = None, main_layer : str = "velosum", extra_layers = None):
+	"""
+		Create a .loom file from STARsolo output
+		Args:
+			  indir (str):	              path to STARsolo output folder (the one that contains 'Solo.out')
+			  outfile (str):              path and name of the new loom file
+			  sample_id (str):            sample_id (typically 10Xxxx_x)
+			  cell_filter (str):          'emptydrops', 'star', 'none', 'combine', 'Gene', 'GeneFull', 'GeneFull_ExonOverIntron', 'GeneFull_Ex50pAS'
+			                              emptydrops: with parameters expected_n_cells, min_total_umis, and ambient_pthreshold
+			                              star: uses the barcodes from 'Solo.out/Velocyto/filtered' subdir of indir (Same as 'Gene')
+			                              combine: combines cells from all extra_layers and fills layer's non-valid cells with zeroes
+			                              Gene... : uses the cells from that Solo.out filtered subdir. Will be added to extra_layers.
+			                              none: all cells are kept - Lots of cells!
+			  sample_metadata_file (str): file of sample metadata or path to sqlite3 database. Used for gene annotation.
+			  gtf_file (str):             path to gtf file used by STARsolo. Used for cell annotations.
+			  main_layer(str):            The STAR output matrix to use for the main loom layer (e.g. 'Gene', or 'GeneFull').
+			                              'velosum' adds up spliced, unspliced, and ambiguous.
+			  extra_layers(list):         Names of additional layers to add in data (e.g. 'GeneFull').
+		Returns:
+		           nothing
+
+	"""
+	gtypes = extra_layers if extra_layers else []
+	if cell_filter in ("Gene", "GeneFull", "GeneFull_ExonOverIntron", "GeneFull_Ex50pAS") and cell_filter not in gtypes:
+		gtypes.append(cell_filter)
+	if main_layer in ("Gene", "GeneFull", "GeneFull_ExonOverIntron", "GeneFull_Ex50pAS") and main_layer not in gtypes:
+		gtypes.append(main_layer)
+	subdir = "raw" if cell_filter == "none" else "filtered"
+	velodir = os.path.join(indir, "Solo.out", "Velocyto", subdir)
+	accessions = [ l.split('\t')[0] for l in open(os.path.join(velodir, "features.tsv")).readlines() ]
+	genes = [ l.split('\t')[1] for l in open(os.path.join(velodir, "features.tsv")).readlines() ]
+	gtype2bcs = {}
+	n_genes = len(accessions)
+	layers = {}
+	for gtype in gtypes:
+		mtxpath = os.path.join(indir, "Solo.out", gtype, subdir, "matrix.mtx")
+		bcspath = os.path.join(indir, "Solo.out", gtype, subdir, "barcodes.tsv")
+		if os.path.exists(mtxpath) and os.path.exists(bcspath):
+			bcs = [ l.rstrip() for l in open(bcspath).readlines() ]
+			gtype2bcs[gtype] = bcs
+			mtx = np.loadtxt(mtxpath, skiprows=3, delimiter=' ')
+			layers[gtype] = sparse.csc_matrix((mtx[:,2], (mtx[:,0]-1, mtx[:,1]-1)), shape = (n_genes, len(bcs))).todense()
+	velobcspath = os.path.join(velodir, "barcodes.tsv")
+	velobcs = [ l.rstrip() for l in open(velobcspath).readlines() ]
+	gtype2bcs['spliced'] = gtype2bcs['unspliced'] = gtype2bcs['ambiguous'] = velobcs
+	velomtxshape = (n_genes, len(velobcs))
+	common_mtx = os.path.join(velodir, "matrix.mtx")
+	if os.path.exists(common_mtx):
+		mtx = np.loadtxt(common_mtx, skiprows=3, delimiter=' ')
+		layers['spliced'] = allspliced = sparse.csc_matrix((mtx[:,2], (mtx[:,0]-1, mtx[:,1]-1)), shape = velomtxshape).todense()
+		layers['unspliced'] = sparse.csc_matrix((mtx[:,3], (mtx[:,0]-1, mtx[:,1]-1)), shape = velomtxshape).todense()
+		layers['ambiguous'] = sparse.csc_matrix((mtx[:,4], (mtx[:,0]-1, mtx[:,1]-1)), shape = velomtxshape).todense()
+	else: # STAR >= 2.7.9
+		for vtype in ('spliced', 'unspliced', 'ambiguous'):
+			mtx = np.loadtxt(os.path.join(velodir, vtype + ".mtx"), skiprows=3, delimiter=' ')
+			layers[vtype] = sparse.csc_matrix((mtx[:,2], (mtx[:,0]-1, mtx[:,1]-1)), shape = velomtxshape).todense()
+	if cell_filter == "emptydrops":
+		allspliced = layers['spliced']
+		spliced_total_umis = np.array(allspliced.sum(axis=0))[0]
+		ambient_umis, ambient_pvalue = call_cells(allspliced.tocsc(), expected_n_cells)
+		valid_cell_idxs = (ambient_pvalue <= ambient_pthreshold) | (spliced_total_umis >= min_total_umis)
+		valid_pvalue = ambient_pvalue[valid_cell_idxs]
+		valid_bcs = gtype2bcs['spliced'][valid_cell_idxs]
+	elif cell_filter == "combine":
+		all_bcs = set()
+		for bcs in gtype2bcs.values():
+			all_bcs.update(bcs)
+		valid_bcs = all_bcs
+	elif cell_filter in gtypes:
+		valid_bcs = gtype2bcs[cell_filter]
+	elif cell_filter !="none":
+		if cell_filter != "star":
+			print (f"Unknown cell_filter type: {cell_filter} - using cells from Solo.out/Velocyto/filtered/")
+		valid_bcs = velobcs # gtype2bcs["Gene"]
+
+	valid_bcs = list(valid_bcs)
+	n_valid_cells = len(valid_bcs)
+	new_shape = (n_genes, n_valid_cells)
+	cellids = np.array([f"{sample_id}:{v_bc}" for v_bc in valid_bcs])
+	ca = { "CellID": cellids }
+	for gtype in layers:
+		valid_cells_layer = np.full(new_shape, 0, layers[gtype].dtype)
+		ca_valid = np.full(n_valid_cells, 0)
+		#print (valid_cells_layer.shape, layers[gtype].shape)
+		for fromidx, bc in enumerate(gtype2bcs[gtype]):
+			try:
+				toidx = valid_bcs.index(bc)
+				#print (fromidx, toidx, bc, layers[gtype].shape, valid_cells_layer.shape)
+				valid_cells_layer[:,toidx] = layers[gtype][:,fromidx].reshape(n_genes,)
+				ca_valid[toidx] = 1
+			except ValueError:
+				pass #print ("Err:", fromidx, bc, len(valid_bcs))
+		layers[gtype] = valid_cells_layer
+		#print (gtype, valid_cells_layer.shape, valid_cells_layer.sum())
+		ca["Valid_" + gtype] = ca_valid
+		#print ("ca", gtype, ca_valid.shape)
+	if main_layer == "velosum":
+		layers[''] = layers['spliced'] + layers['unspliced'] + layers['ambiguous']
+	else:
+		layers[''] = layers[main_layer]
+		del layers[main_layer]
+	total_umis = layers[''].sum(axis=0)
+	ca["BarcodeTotalUMIs"] = total_umis
+	if cell_filter == "emptydrops":
+		ca["AmbientPValue"] = valid_pvalue
+		ca["AmbientUMIs"] = np.full(n_valid_cells, ambient_umis)
+	if sample_metadata_file:
+		sample_metadata = load_sample_metadata(sample_metadata_file, sample_id)
+		for key, value in sample_metadata.items():
+			ca[key] = np.full(n_valid_cells, value)
+	if gtf_file:
+		ra = make_row_attrs_from_gene_metadata(gtf_file, accessions)
+	else:
+		ra = { "Gene": np.array(genes), "Accession": np.array(accessions) }
+	create(filename=outfile, layers=layers, row_attrs=ra, col_attrs=ca)
+
 def create_from_kallistobus(out_file: str, in_dir: str, tr2g_file: str, whitelist_file: str, file_attrs: Dict[str, str] = None, layers: Dict[str, str] = None):
 	"""
 	Create a loom file from a kallisto-bus output folder.
@@ -1387,3 +1513,4 @@ def connect(filename: str, mode: str = 'r+', *, validate: bool = True, spec_vers
 		Note: if validation is requested, an exception is raised if validation fails.
 	"""
 	return LoomConnection(filename, mode, validate=validate)
+
